# Copyright 2016 Commerce Technologies, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from kerb_sts.awsrole import AWSRole


class TestAWSRoleCreation(unittest.TestCase):
    def test_with_role_as_none(self):
        is_valid = AWSRole.is_valid(None)
        self.assertFalse(is_valid)

    def test_with_empty_string(self):
        is_valid = AWSRole.is_valid('')
        self.assertFalse(is_valid)

    def test_with_malformed_role_string(self):
        is_valid = AWSRole.is_valid('arn_noseparator_provider')
        self.assertFalse(is_valid)

    def test_with_missing_arn_string(self):
        is_valid = AWSRole.is_valid(',provider')
        self.assertFalse(is_valid)

    def test_with_missing_provider_string(self):
        is_valid = AWSRole.is_valid('arn/role,')
        self.assertFalse(is_valid)

    def test_with_valid_strings(self):
        is_valid = AWSRole.is_valid('arn/role,provider')
        self.assertTrue(is_valid)

    def test_parsed_role(self):
        role = AWSRole('arn/role,provider')
        self.assertEqual(role.arn, 'arn/role')
        self.assertEqual(role.provider, 'provider')

    def test_with_valid_strings(self):
        role = AWSRole('arn/role,provider')
        self.assertEqual(role.name, 'role')


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Commerce Technologies, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os


def _get_default_config_filename():
    """
    Returns the path of the default Kerb-STS configuration file.
    This is the path that Kerb will look for configuration and
    the location it will write configuration to when the --configure
    parameter is passed in.
    """
    return os.path.expanduser('~/.kerb-sts/config.json')


class Config:
    """
    The Config object stores connection configuration for
    the tool.
    """
    IDP_URL_KEY = 'idp_url'
    ADFS_URL_KEY = 'adfs_url'
    KERB_HOSTNAME_KEY = 'kerb_hostname'
    REGION_KEY = 'region'
    PREFERRED_AUTH_TYPE = 'preferred_auth_type'

    def __init__(self, idp_url, kerb_hostname, region, preferred_auth_type):
        self.idp_url = idp_url
        self.kerb_hostname = kerb_hostname
        self.region = region
        self.preferred_auth_type = preferred_auth_type

    def save(self, filename=_get_default_config_filename()):
        """
        Serializes the config object into the default config location.
        :param filename: where to save the Config to
        """
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        dictionary = {
            Config.IDP_URL_KEY: self.idp_url,
            Config.KERB_HOSTNAME_KEY: self.kerb_hostname,
            Config.REGION_KEY: self.region,
            Config.PREFERRED_AUTH_TYPE: self.preferred_auth_type
        }
        with open(filename, 'w') as f:
            json.dump(dictionary, f)
            logging.info("config file saved to {}".format(filename))

    def is_valid(self):
        """
        Validates that the Config object is valid (non-null non-empty values).
        :return: True if the Config is valid
        """
        return (self.idp_url is not None and self.idp_url is not '' and
                self.region is not None and self.region is not '')

    @staticmethod
    def load(filename=_get_default_config_filename()):
        """
        A factory method for returning a Config object
        that has been deserialized from the given filename.
        :param filename: the path to the config file
        :return: a Config object constructed from the filename's contents
        """
        config = Config(idp_url='', kerb_hostname='', region='', preferred_auth_type='')
        try:
            with open(filename) as f:
                config_json = json.loads(f.read())
                if config_json:
                    if Config.IDP_URL_KEY in config_json:
                        config.idp_url = str(config_json[Config.IDP_URL_KEY])
                    elif Config.ADFS_URL_KEY in config_json:
                        config.idp_url = str(config_json[Config.ADFS_URL_KEY])

                    if Config.KERB_HOSTNAME_KEY in config_json:
                        config.kerb_hostname = str(config_json[Config.KERB_HOSTNAME_KEY])

                    if Config.REGION_KEY in config_json:
                        config.region = str(config_json[Config.REGION_KEY])

                    if Config.PREFERRED_AUTH_TYPE in config_json:
                        config.preferred_auth_type = str(config_json[Config.PREFERRED_AUTH_TYPE])
        except IOError:
            logging.warning("could not read config file {}".format(filename))
        return config



# Copyright 2016 Commerce Technologies, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pexpect
import subprocess

from requests_kerberos import HTTPKerberosAuth, OPTIONAL
from requests_ntlm import HttpNtlmAuth


class Authenticator(object):
    """
    The Authenticator class is an abstract class for shared functionality
    between the possible authentication types.
    """

    def get_auth_handler(self, session):
        raise NotImplementedError("override this method")

    @staticmethod
    def get_auth_type():
        raise NotImplementedError("override this method")


class KerberosAuthenticator(Authenticator):
    """
    The KerberosAuthenticator uses the local Kerberos install to
    authenticate a user who's machine is logged into to the Domain.
    """
    AUTH_TYPE = 'kerberos'

    def __init__(self, kerb_hostname=None, username=None, password=None, domain=None):
        self.kerb_hostname = kerb_hostname if kerb_hostname else None

        if username:
            KerberosAuthenticator.__generate_kerberos_ticket(username, password, domain)
        else:
            KerberosAuthenticator.__generate_kerberos_ticket_using_credentials_cache()

    @staticmethod
    def __generate_kerberos_ticket(username, password, domain):
        try:
            principal = '{}@{}'.format(username, domain)
            kinit = pexpect.spawn('kinit {}'.format(principal))
            kinit.sendline(password)
            kinit.expect(pexpect.EOF)
            kinit.close()
            if kinit.exitstatus is not 0:
                raise Exception('kinit failed for principal {}'.format(principal))
        except Exception as ex:
            raise ex

    @staticmethod
    def __generate_kerberos_ticket_using_credentials_cache():
        # Windows does not have support for `klist`. Assume
        # Windows users have a valid Kerberos ticket.
        if os.name != 'nt':
            try:
                subprocess.check_output(['klist', '-s'])
            except subprocess.CalledProcessError:
                logging.info('no kerberos ticket found. running kinit')
                try:
                    subprocess.check_output(['kinit'])
                except subprocess.CalledProcessError as err:
                    logging.error('failed to generate a kerberos ticket')
                    raise err

    def get_auth_handler(self, session):
        return HTTPKerberosAuth(mutual_authentication=OPTIONAL, hostname_override=self.kerb_hostname)

    @staticmethod
    def get_auth_type():
        return KerberosAuthenticator.AUTH_TYPE


class NtlmAuthenticator(Authenticator):
    """
    The NtlmAuthenticator authenticates users with basic credentials
    and a domain.
    """
    AUTH_TYPE = 'ntlm'

    def __init__(self, username, password, domain):
        self.username = username
        self.password = password
        self.domain = domain

    def get_auth_handler(self, session):
        return HttpNtlmAuth(
            "{}\\{}".format(self.domain, self.username),
            self.password,
            session
        )

    @staticmethod
    def get_auth_type():
        return NtlmAuthenticator.AUTH_TYPE


class KeytabAuthenticator(Authenticator):
    """
    The KeytabAuthenticator allows users to sign keytabs with
    their password and then request valid Kerberos tickets
    with those keytabs without requiring password input.
    """

    def __init__(self, kerb_hostname, username, keytab, domain):
        self.kerb_hostname = kerb_hostname
        self.username = username
        self.keytab = keytab
        self.domain = domain

        if os.name == 'nt':
            raise Exception("keytab is not supported on Windows!")

        # Setup a temporary credentials cache
        credentials_cache = os.path.join(os.getcwd(), 'credentials_cache')
        os.environ['KRB5CCNAME'] = credentials_cache

        # Call kinit to generate a Kerberos ticket for the given username and keytab
        try:
            subprocess.check_output(['kinit', '-c', credentials_cache, '-kt', self.keytab,
                                     "{}@{}".format(self.username, self.domain)])
        except subprocess.CalledProcessError:
            raise Exception("could not generate a valid ticket for the given keytab")

    def get_auth_handler(self, session):
        return HTTPKerberosAuth(mutual_authentication=OPTIONAL, hostname_override=self.kerb_hostname)

    @staticmethod
    def get_auth_type():
        return 'keytab'




# Copyright 2016 Commerce Technologies, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import os
import sys
import time

from six.moves import input

from kerb_sts.config import Config
from kerb_sts.handler import KerberosHandler
from kerb_sts.auth import KerberosAuthenticator, NtlmAuthenticator, KeytabAuthenticator

DEFAULT_REGION = 'us-east-1'


def _get_default_credentials_filename():
    """
    Returns the path to the default AWS credentials file. This
    is where the temporary access keys will be stored unless
    otherwised specified by the user.
    :return: The path to the default credentials file
    """

    return os.path.expanduser('~/.aws/credentials')


def _get_default_config_filename():
    """
    Returns the path to the default AWS credentials file. This
    is where the temporary access keys will be stored unless
    otherwised specified by the user.
    :return: The path to the default credentials file
    """

    return os.path.expanduser('~/.aws/config')


def _get_options():
    """
    Parses the command line options.
    :return: an options object
    """
    parser = argparse.ArgumentParser(description="Generates 1 hour temporary AWS IAM credentials.")
    parser.add_argument('--idp_url', help="IdP domain name", dest='idp_url')
    parser.add_argument('--kerb_hostname', help="Kerberos hostname (defaults to IdP domain)", dest='kerb_hostname')
    parser.add_argument('-c', '--credentials_file', help="AWSCLI credentials file (defaults ~/.aws/credentials)",
                        dest='credentials_file', default=_get_default_credentials_filename())
    parser.add_argument('-config_file', help="AWSCLI config file (defaults ~/.aws/config)",
                        dest='config_file', default=_get_default_config_filename())
    parser.add_argument('--configure', help="Sets up Kerb-STS by generating a config file at ~/.kerb-sts/config",
                        dest='configure', action='store_true', default=False)
    parser.add_argument('--daemon', help="Run as a daemon. This will auto-renew credentials every half hour",
                        dest='daemon', action='store_true', default=False)
    parser.add_argument('-r', '--default_role', help="Name of the Role to use as the default",
                        dest='default_role', default=None)
    parser.add_argument('-o', '--only_role', help="Name of the Role to use as the default (no other roles will be processed)",
                        dest='only_role', default=None)
    parser.add_argument('-d', '--domain', help="AD Domain if using a Kerberos keytab or NTLM auth. Requires a username and password/keytab",
                        dest='domain', default=None)
    parser.add_argument('--keytab', help="The Kerberos keytab file. Requires a username and domain",
                        dest='keytab', default=None)
    parser.add_argument('--list', help="List the available roles",
                        dest='list', action='store_true', default=False)
    parser.add_argument('-p', '--password', help="AD Password if generating a temporary Kerberos token. Requires a username and domain",
                        dest='password', default=None)
    parser.add_argument('--refresh', help="Time to wait (minutes) between refreshing the tokens.",
                        dest='refresh', default=30)
    parser.add_argument('--region', help="AWS Region for STS (defaults to {})".format(DEFAULT_REGION),
                        dest='region', default=DEFAULT_REGION)
    parser.add_argument('-u', '--username', help="AD Username if generating a temporary Kerberos token. Requires a domain and password/keytab",
                        dest='username', default=None)
    parser.add_argument('-v', '--verbose', help="Turns on debug logging",
                        dest="verbose", action='store_true', default=None)
    return parser.parse_args()


def _setup_logging(options):
    """
    Sets up logging based on some command line options.
    :param options: options parsed from the command line
    """
    if options.verbose:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p', level=logging_level)
    logging.debug("logging level set to {}".format(logging_level))


def _configure():
    """
    Generates a configuration file for subsequent runs to consume.
    """
    idp_url = input('IdP AWS sign in URL: ')
    kerb_hostname = input('Kerberos hostname (default: IdP AWS sign in URL domain): ')

    region = input('AWS region (default: {}): '.format(DEFAULT_REGION))
    if region == '':
        region = DEFAULT_REGION

    preferred_auth_type = input('Preferred auth type (default: ntlm): ')
    valid_preferred_auth_types = [KerberosAuthenticator.AUTH_TYPE, NtlmAuthenticator.AUTH_TYPE]
    if preferred_auth_type == '':
        preferred_auth_type = NtlmAuthenticator.AUTH_TYPE
    elif preferred_auth_type not in valid_preferred_auth_types:
        raise Exception('invalid preferred auth type; acceptable types: {}'.format(valid_preferred_auth_types))

    config = Config(
        idp_url=idp_url,
        region=region,
        kerb_hostname=kerb_hostname,
        preferred_auth_type=preferred_auth_type
    )
    config.save()


def _setup_config(options):
    """
    Creates a config object and overrides any values
    provided on the command line.
    :param options: parsed command line options
    :return: the Config object
    """
    config = Config.load()

    if options.idp_url:
        config.idp_url = options.idp_url
    logging.debug('IdP url set to {}'.format(config.idp_url))

    if options.region:
        config.region = options.region
    logging.debug('region set to {}'.format(config.region))

    return config


def _setup_authenticator(options, config):
    """
    Creates an Authenticator object based on
    what credential information was passed as arguments.
    :param options: pasred command line options
    :return: an Authenticator object
    """
    if options.username and options.domain:
        if options.password:
            if config.preferred_auth_type == KerberosAuthenticator.AUTH_TYPE:
                authenticator = KerberosAuthenticator(
                    kerb_hostname=config.kerb_hostname,
                    username=options.username,
                    password=options.password,
                    domain=options.domain
                )
            elif config.preferred_auth_type == NtlmAuthenticator.AUTH_TYPE or not config.preferred_auth_type:
                authenticator = NtlmAuthenticator(
                    username=options.username,
                    password=options.password,
                    domain=options.domain
                )
            else:
                raise Exception('invalid preferred auth type {}'.format(config.preferred_auth_type))
        elif options.keytab:
            authenticator = KeytabAuthenticator(
                kerb_hostname=config.kerb_hostname,
                username=options.username,
                keytab=options.keytab,
                domain=options.domain
            )
        else:
            raise Exception("username and domain provided but no password or keytab was given")
    elif options.username or options.domain:
        raise Exception("both username and domain are required for ntlm or keytab authentication")
    else:
        authenticator = KerberosAuthenticator(config.kerb_hostname)
    return authenticator


def _generate_tokens(options, config, authenticator):
    """
    Generates a set of AWS IAM credentials for each available role
    for the principal.
    :param options: the parsed command line arguments
    :param config: the kerb configuration
    :param authenticator: the Authenticator object used to handle IdP authentication
    """
    logging.info("--------------------------------")
    if options.list:
        logging.info("    listing available roles     ")
    else:
        logging.info("       generating tokens        ")
    logging.info("--------------------------------")

    logging.info("region: {}".format(config.region))
    logging.info("url: {}".format(config.idp_url))
    logging.info("credentials file: {}".format(options.credentials_file))
    if options.only_role:
        logging.info("only role: {}".format(options.only_role))
        options.default_role = options.only_role
    elif options.default_role:
        logging.info("default role: {}".format(options.default_role))
    logging.info("auth type: {}".format(authenticator.get_auth_type()))

    h = KerberosHandler()
    h.handle_sts_by_kerberos(config.region, config.idp_url, options.credentials_file, options.config_file,
                             options.only_role, options.default_role, options.list, authenticator)
    logging.info("--------------------------------")


def main():
    """
    Application entrypoint. Parses the command line options, sets up the configuration for the token
    generation, and then starts generating token(s) for the given options.
    :return:
    """
    options = _get_options()

    _setup_logging(options)

    if options.configure:
        _configure()
        sys.exit(0)

    config = _setup_config(options)
    if not config.is_valid():
        logging.error(
            "invalid configuration. please run --configure to generate a config file or supply a valid IdP URL")
        sys.exit(1)

    try:
        authenticator = _setup_authenticator(options, config)
    except Exception as ex:
        logging.error(ex)
        sys.exit(1)

    while True:
        try:
            _generate_tokens(options, config, authenticator)
        except Exception as ex:
            logging.error(ex)
            sys.exit(1)

        if not options.daemon:
            break
        logging.info('')
        time.sleep(60 * options.refresh)


if __name__ == "__main__":
    main()
